# This function performs one repetition of the five times repeated 
# 5-fold stratified cross-validation on a specific data set using
# one of the fives compared methods.
#
# It takes the whole number 'iter', which corresponds to the iter-th line 
# of 'scenariogrid', which contains the necessary information
# on the iter-th setting.

evaluatesetting <- function(iter) {
  
  # Obtain information for the iter-th setting:
  
  dataid <- scenariogrid$dataid[iter]
  method <- scenariogrid$method[iter]
  cvind <- scenariogrid$cvind[iter]
  seed <- scenariogrid$seed[iter]
  settingid <- scenariogrid$settingid[iter]
  
  
  # Set seed:
  
  set.seed(seed)
  
  
  # Load data set:
  
  alldatasets <- list.files("./InteractionForests/Data/Datasets")
  datasetname <- grep(paste("_id", dataid, ".Rda", sep=""), alldatasets, value=TRUE)
  
  load(paste("./InteractionForests/Data/Datasets/", datasetname, sep=""))
  
  
  # Make CV division:
  
  Kcv <- 5
  
  cvdiv <- makeCVdiv(data=dataset, yname="Ytarget", ncv=Kcv)
  
  
  # Perform cross-validation: 
  
  res <- list()
  
  for(i in 1:Kcv) {
    
    datatrain <- dataset[cvdiv!=i,]
    datatest <- dataset[cvdiv==i,]
    
    if(method=="randomforest") {
      
      res[[i]] <- trainandpredict(datatrain, datatest, trainfun=trainRandomForest, predictfun=predictRandomForest)
      
    }
    
    if(method=="interactionforest") {
      
      res[[i]] <- trainandpredict(datatrain, datatest, trainfun=trainInteractionForest, predictfun=predictInteractionForest)
      
    }
    
    if(method=="rotationforest") {
      
      res[[i]] <- trainandpredict(datatrain, datatest, trainfun=trainRotationForest, predictfun=predictRotationForest)
      
    }
    
    if(method=="obliqueforest") {
      
      res[[i]] <- trainandpredict(datatrain, datatest, trainfun=trainObliqueForest, predictfun=predictObliqueForest)
      
    }
    
    if(method=="canonicalforest") {
      
      res[[i]] <- trainandpredict(datatrain, datatest, trainfun=trainCanForest, predictfun=predictCanForest)
      
    }
    
    
  }
  
  # Save results:
  
  save(res, file=paste("./InteractionForests/Results/LargeStudy_Rda_files/res_", settingid, ".Rda", sep=""))

}









# Wrapper function that fits a prediction rule on a training
# data set and performs performance evaluation on a test
# data set.

# Input parameters:

# datatrain  - training data set
# datatest   - test data set
# trainfun   - function that fits the prediction rule
# predictfun - prediction function

trainandpredict <- function(datatrain, datatest, trainfun, predictfun) {
  
  # Construct forest:
  model <- trainfun(dependent.variable.name="Ytarget", data=datatrain)
  
  # Perform prediction:
  preds <- predictfun(model, dependent.variable.name="Ytarget", newdata=datatest)
  probpreds <- preds$probpreds
  classpreds <- preds$classpreds
  
  # Calculate AUC values:
  auctest <- measureAUC(probabilities=probpreds, truth=datatest$Ytarget, 
                        negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Calculate ACC values:
  acctest <- measureACC(truth=datatest$Ytarget, response=classpreds)
  
  # Calculate Brier values:
  briertest <- measureBrier(probabilities=probpreds, truth=datatest$Ytarget, 
                            negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Return results:
  return(list(acctest=acctest, auctest=auctest, briertest=briertest))
  
}










# Function that constructs a Rotation Forest:

trainRotationForest <- function(dependent.variable.name, data, ...) {
  
  require("fastDummies")
  require("rotationForest")
  require("nnet")
  
  # Dummy code factor variables:
  factorvars <- setdiff(names(data)[sapply(data, is.factor)], dependent.variable.name)
  if (length(factorvars) > 0) {
    datatrans <- fastDummies::dummy_cols(data, select_columns = factorvars, remove_selected_columns=TRUE, remove_first_dummy=TRUE)
  } else {
    datatrans <- data 
  }
  
  # Apply rotationForest:
  yvar <- as.factor(as.numeric(datatrans[,dependent.variable.name])-1)
  model <- rotationForest::rotationForest(x=datatrans[,names(datatrans)!=dependent.variable.name], y=yvar, ...)
  
  return(model)
  
}



# This is an altered version of the function predict.rotationForest from
# the rotationForest package.
# In the original version of the package, the
# command "newdata <- data.frame(sapply(newdata, as.numeric, simplify = FALSE))"
# changed the column names for some datasets with specific characters
# in the column names. Therefore the fixed version, safes the column
# names before the above command and reassigns them after the command.

library("rotationForest")

mypredictrotationForest <- function (object, newdata, all = FALSE, ...) 
{
  
  colnamessafe <- colnames(newdata)    ###### CHANGE: Safe column names.
  newdata <- data.frame(sapply(newdata, as.numeric, simplify = FALSE))
  colnames(newdata) <- colnamessafe    ###### CHANGE: Reassign the safed column names.
  if (!identical(colnames(newdata), object$columnnames)) 
    stop("Variable names and/or order of variables in newdata is not identical to training set. Please check if variables are exactly the same in both sets.")
  predicted <- matrix(NA, nrow = nrow(newdata), ncol = length(object$models))
  for (i in 1:length(object$models)) {
    final <- data.frame(as.matrix(newdata) %*% as.matrix(object$loadings[[i]]))
    predicted[, i] <- predict(object$models[[i]], final, 
                              type = "prob")[, 2]
  }
  if (all) {
    predicted
  }
  else {
    rowMeans(predicted)
  }
}

environment(mypredictrotationForest) <- environment(rotationForest)



# Prediction function for Rotation Forest:

predictRotationForest <- function(model, dependent.variable.name, newdata) {
  
  require("fastDummies")
  require("rotationForest")
  
  # Dummy code factor variables:
  factorvars <- setdiff(names(newdata)[sapply(newdata, is.factor)], dependent.variable.name)
  if (length(factorvars) > 0) {
    newdatatrans <- fastDummies::dummy_cols(newdata, select_columns = factorvars, remove_selected_columns=TRUE, remove_first_dummy=TRUE)
  } else {
    newdatatrans <- newdata 
  }
  
  # Perform prediction:
  probpreds <- mypredictrotationForest(object=model, newdata=newdatatrans[,names(newdatatrans)!=dependent.variable.name])
  classpreds <- factor(levels(newdata[,dependent.variable.name])[as.numeric(probpreds > 0.5)+1], levels=levels(newdata[,dependent.variable.name]))
  
  return(list(probpreds=probpreds, classpreds=classpreds))
  
}












# This is an corrected version of the function obliqueRF from
# the obliqueRF package.
# The following line
#  if (!all(y == 0 | y == 1)) {
# was replaced by the following line
#  if (!(all(y == 0 | y == 1) & !is.factor(y))) {

library("obliqueRF")

myobliqueRF <- function (x, y, x.test = NULL, y.test = NULL, mtry = NULL, ntree = 100, 
                         training_method = "ridge", bImportance = F, bProximity = F, 
                         verbose = F, ...) 
{
  num_split_vars <- mtry
  do_oob <- TRUE
  if (verbose) 
    cat("obliqueRF: WARNING: status messages enabled; set \"verbose\" to false to disable\n")
  impStruct <- list(bImportance = bImportance, importance_pvalue = 0, 
                    importance_weight = "")
  if (bImportance) {
    impStruct$importance_pvalue <- 0.05
    impStruct$importance_weight <- "none"
    do_oob <- FALSE
    if (bProximity) {
      cat("obliqueRF: WARNING: importance and proximity cannot be calculated at the same time. Will set 'bProximity = FALSE' and continue.\n")
      bProximity = FALSE
    }
    if (!(training_method == "log")) {
      cat("obliqueRF: WARNING: importance requires training_method=\"log\". Adjusting training_method...\n")
      training_method = "log"
    }
  }
  node_train_method = NULL
  if (training_method == "ridge") {
    node_train_method = ridge_svd_wrapper
    cat("obliqueRF: using fast ridge regression with SVD as node model,\n does not scale data at the node,\n tests regularization parameter \"lambda\" for 10^(c(-5:5)).\n In case of problems with the SVD variant try training_method=\"ridge_slow\" for separate ridge regressions (slower).\n")
  }
  if (training_method == "ridge_slow") {
    node_train_method = ridge_slow_wrapper
    cat("obliqueRF: using slow ridge regression as node model,\n does not scale data at the node,\n tests regularization parameter \"lambda\" for 10^(c(-5:5)) with separate explicit ridge regressions.\n")
  }
  if (training_method == "log") {
    node_train_method = log_wrapper
    cat("obliqueRF: using logistic regression as node model.\n")
  }
  if (training_method == "svm") {
    node_train_method = svm_wrapper
    cat("obliqueRF: using linear SVM as node model,\n scales data at the node,\n tests regularization parameter \"cost\" for 10^(c(-3,-1,1)).\n")
  }
  if (training_method == "pls") {
    node_train_method = pls_wrapper
    cat("obliqueRF: using partial least squares regression as node model,\n does not scale data at the node,\n optimizes regularization parameter \"ncomp\" (number of components) for 1:min(20, mtry).\n")
  }
  if (training_method == "rnd") {
    node_train_method = rnd_wrapper
    cat("obliqueRF: using a random hyperplane as node model,\n scales data to unit variance at the node,\n coefficients of random subspace projection are chosen from a normal distribution with unit variance.\n")
  }
  if (is.null(node_train_method)) {
    stop("obliqueRF: the training method is not valid, choose one of the following:\n  ridge, ridge_slow, log, svm, pls or rnd.\n  See documentation for more details. Aborting")
  }
  node_pred_method = predict
  node_impurity_method = gini_impurity_wrapper
  if (any(is.nan(as.vector(x))) | any(is.infinite(as.vector(x))) | 
      any(is.nan(as.vector(y))) | any(is.infinite(as.vector(y)))) 
    stop("obliqueRF: Training data contains NaNs or Inf values. Please correct and try again. Aborting...")
  if (is.null(y.test)) {
    cat("obliqueRF: no test set defined. Will use training data.\n")
    x.test <- x
    y.test <- y
  }
  if (sum(is.nan(as.vector(x.test))) > 0 | sum(is.infinite(as.vector(x.test))) > 
      0 | sum(is.nan(as.vector(y.test))) > 0 | sum(is.infinite(as.vector(y.test))) > 
      0) 
    stop("obliqueRF: Test data contains NaNs or Inf values. Please correct and try again. Aborting...")
  clnames <- sort(unique(paste(y)))
  # if (!all(y == 0 | y == 1)) {
  if (!(all(y == 0 | y == 1) & !is.factor(y))) {
    yout <- rep(0, length(y))
    ytestout <- rep(0, length(y.test))
    yout[y == clnames[1]] <- 0
    ytestout[y.test == clnames[1]] <- 0
    yout[y == clnames[2]] <- 1
    ytestout[y.test == clnames[2]] <- 1
    y <- yout
    y.test <- ytestout
  }
  num_vars = ncol(x)
  num_samples = nrow(x)
  if (length(y) != num_samples) 
    stop("obliqueRF: WARNING: number of samples and length of y does not match.")
  test_set <- cbind(as.matrix(x.test), y.test)
  d <- cbind(as.matrix(x), y)
  cl_id = num_vars + 1
  cl <- y
  if (is.null(num_split_vars)) 
    num_split_vars = max(as.integer(sqrt(num_vars)), 2)
  if (num_split_vars > num_vars) {
    stop("obliqueRF: WARNING: mtry > number of variable.")
  }
  num_classes <- length(unique(cl))
  if ((num_classes < 2) || min(cl)) 
    stop("obliqueRF: WARNING: wrong number of classes. Aborting...")
  nr_nodes = 2 * num_samples + 1
  nd_min_size = 1
  oob_glob = rep(0, num_samples)
  oob_class_votes <- matrix(0, num_classes, nrow(d))
  test_set_class_votes <- matrix(0, num_classes, nrow(test_set))
  if (bImportance) {
    imp <- rep(0, ncol(x))
    imp_class_votes <- matrix(0, num_classes, nrow(d))
  }
  oob_prox <- matrix(NA, nrow(d), ntree)
  err <- c()
  err_0 <- c()
  err_1 <- c()
  trees = vector("list", ntree)
  if (verbose) 
    cat("obliqueRF: starting.", date(), "\n")
  for (curr_num_tree in 1:ntree) {
    if (verbose) 
      if ((curr_num_tree%%10) == 1) 
        cat("", curr_num_tree, "out of", 
            ntree, "trees so far.", date(), "\n")
    class_n = rep(0, num_classes)
    k <- sample(1:num_samples, num_samples, replace = TRUE)
    sample_pop = as.numeric(table(factor(k, levels = 1:num_samples)))
    class_n = as.numeric(table(cl[k]))
    dframe <- d
    b_sample = sample_pop
    nr_diff_cases = length(sample_pop > 0)
    n_case = 1:num_samples
    tree <- buildtree(dframe, b_sample, class_n, n_case, 
                      nr_nodes, nd_min_size, nr_diff_cases, num_vars, num_split_vars, 
                      num_classes, node_train_method, node_pred_method, 
                      node_impurity_method, num_samples, impStruct)
    trees[[curr_num_tree]] = tree
    node_classIdx = tree$node_class + 1
    if (do_oob) {
      oob_glob = oob_glob + (sample_pop == 0) * 1
      for (M in which(sample_pop == 0)) {
        curr_node = 1
        for (i in 1:(tree$total_nr_nodes)) {
          if (tree$node_status[curr_node] == -1) {
            oob_class_votes[node_classIdx[curr_node], 
                            M] = oob_class_votes[node_classIdx[curr_node], 
                                                 M] + 1
            oob_prox[M, curr_num_tree] <- curr_node
            break
          }
          dset <- d[M, tree$best_vars[, curr_node]]
          score = sum(tree$training_stats[, curr_node] * 
                        dset)
          if (score < tree$best_split_value[curr_node]) 
            curr_node = tree$treemap[1, curr_node]
          else curr_node = tree$treemap[2, curr_node]
        }
      }
    }
    if (bImportance) {
      for (pi in 1:ncol(tree$signif)) imp[tree$best_vars[, 
                                                         pi]] <- imp[tree$best_vars[, pi]] + (tree$significance[, 
                                                                                                                pi])
    }
    for (M in 1:nrow(test_set)) {
      curr_node = 1
      for (i in 1:(tree$total_nr_nodes)) {
        if (tree$node_status[curr_node] == -1) {
          test_set_class_votes[node_classIdx[curr_node], 
                               M] = test_set_class_votes[node_classIdx[curr_node], 
                                                         M] + 1
          break
        }
        dset <- test_set[M, tree$best_vars[, curr_node]]
        score = sum(tree$training_stats[, curr_node] * 
                      dset)
        if (score < tree$best_split_value[curr_node]) 
          curr_node = tree$treemap[1, curr_node]
        else curr_node = tree$treemap[2, curr_node]
      }
    }
    v <- 1 * (test_set_class_votes[2, ]/sum(test_set_class_votes[, 
                                                                 1]) > 0.5)
    m <- table(factor(y.test, levels = c(0, 1)), factor(v, 
                                                        levels = c(0, 1)))
    err_0[curr_num_tree] <- m[1, 2]/sum(m[1, ])
    err_1[curr_num_tree] <- m[2, 1]/sum(m[2, ])
    err[curr_num_tree] <- (m[1, 2] + m[2, 1])/sum(m)
  }
  test.votes <- test_set_class_votes[2, ]/sum(test_set_class_votes[, 
                                                                   1])
  if (bProximity) {
    oob_counts <- matrix(0, nrow(d), nrow(d))
    for (op in 1:ncol(oob_prox)) {
      wh <- which(!is.na(oob_prox[, op]))
      oob_counts[wh, wh] <- oob_counts[wh, wh] + 1
    }
    node_counts <- matrix(0, nrow(d), nrow(d))
    for (op in 1:ncol(oob_prox)) {
      tabl <- table(oob_prox[, op])
      mults <- as.numeric(names(tabl)[tabl > 1])
      for (mi in mults) {
        wh <- which(oob_prox[, op] == mi)
        node_counts[wh, wh] <- node_counts[wh, wh] + 
          1
      }
    }
    dst <- 1 - node_counts/oob_counts
    diag(dst) <- 0
  }
  oob.votes <- NULL
  thr <- NULL
  if (do_oob) {
    oob.votes <- oob_class_votes[2, ]/(oob_class_votes[1, 
    ] + oob_class_votes[2, ])
    pred <- prediction(oob.votes, y)
    perf <- performance(pred, "acc")
    thr <- perf@x.values[[1]][which.max(perf@y.values[[1]])]
  }
  cl <- match.call()
  cl[[1]] <- as.name("obliqueRF")
  res <- list(call = cl, type = "classification", class_names = clnames, 
              errs = list(err_0 = err_0, err_1 = err_1, err = err), 
              pred = list(pred.test = 1 * (test.votes > 0.5), votes.test = test.votes, 
                          y.test = y.test, votes.oob = oob.votes, y = y), lab = paste("oRF - node:", 
                                                                                      tree$lab), ntree = ntree, mtry = num_split_vars, 
              importance = if (bImportance) imp else NULL, proximity = if (bProximity) dst else NULL, 
              num_classes = num_classes, trees = trees)
  class(res) <- "obliqueRF"
  return(res)
}

environment(myobliqueRF) <- environment(obliqueRF)



# Function that constructs an Oblique Forest:

trainObliqueForest <- function(dependent.variable.name, data, ...) {
  
  require("fastDummies")
  require("obliqueRF")
  
  # Dummy code factor variables:
  factorvars <- setdiff(names(data)[sapply(data, is.factor)], dependent.variable.name)
  if (length(factorvars) > 0) {
    datatrans <- fastDummies::dummy_cols(data, select_columns = factorvars, remove_selected_columns=TRUE, remove_first_dummy=TRUE)
  } else {
    datatrans <- data 
  }
  
  # Construct Oblique Forest:
  model <- myobliqueRF(x=as.matrix(datatrans[,names(datatrans)!=dependent.variable.name]), y=datatrans[,dependent.variable.name], ...)
  
  return(model)
  
}




# Prediction function for Oblique Forest:

predictObliqueForest <- function(model, dependent.variable.name, newdata) {
  
  require("fastDummies")
  require("obliqueRF")
  
  # Dummy code factor variables:
  factorvars <- setdiff(names(newdata)[sapply(newdata, is.factor)], dependent.variable.name)
  if (length(factorvars) > 0) {
    newdatatrans <- fastDummies::dummy_cols(newdata, select_columns = factorvars, remove_selected_columns=TRUE, remove_first_dummy=TRUE)
  } else {
    newdatatrans <- newdata 
  }
  
  # Perform prediction:
  probpreds <- predict(object=model, newdata=as.matrix(newdatatrans[,names(newdatatrans)!=dependent.variable.name]), type="prob")[,levels(newdata[,dependent.variable.name])[2]]
  classpreds <- factor(predict(object=model, newdata=as.matrix(newdatatrans[,names(newdatatrans)!=dependent.variable.name]), type="response"), levels=levels(newdata[,dependent.variable.name]))
  
  return(list(probpreds=probpreds, classpreds=classpreds))
  
}

















# Function that constructs a Canonical Correlation Forest:

trainCanForest <- function(dependent.variable.name, data, ...) {
  
  
  # This is an altered version of the function setupLeaf from
  # the ccf package.
  # The original version of the package was not able to deal
  # with situations, in which both classes in leafs were
  # represented by the same numbers of observations.
  # In the altered version, one of the two classes is
  # assigned randomly in these situations.
  
  mysetupLeaf <- function (y) 
  {
    if (is.matrix(y)) {
      countsNode <- colSums(y)
      classIndex <- nnet::which.is.max(countsNode)
      #### maxCounts <- max(countsNode)
      #### equalMaxCounts <- maxCounts == countsNode
      #### if (sum(equalMaxCounts) == 1) {
      ####     classIndex <- which(equalMaxCounts)
      #### }
      #### else {
      ####     TODO("TODO multiple max tie breaking")
      #### }
    }
    else {
      countsNode <- sum(y)
      classIndex <- 1
    }
    return(list(isLeaf = TRUE, classIndex = classIndex, trainingCounts = countsNode))
  }
  
  library("ccf")
  library("R.utils")
  reassignInPackage(name="setupLeaf", pkgName="ccf", value=mysetupLeaf)
  
  
  
  # This is an altered version of the function canonical_correlation_tree from
  # the ccf package.
  
  # The following issues were fixed here:
  # 1. In situations in which a tree did not involve at least one split, the class
  #    "canonical_correlation_tree" was not assigned to the tree, which lead to an error
  #    when applying the prediction function, because the latter needs the
  #    trees to have class "canonical_correlation_tree".
  # 2. For nodes, which featured only two observations ("if (nrow(X) == 2) {"),
  #    which did not feature the same class, it was possible that the
  #    training of the tree ran into an infinite loop.
  
  mycanonical_correlation_tree <- function (X, Y, depth = 0, minPointsForSplit = 2, maxDepthSplit = Inf, 
                                            xVariationTolerance = 1e-10, projectionBootstrap = FALSE, 
                                            ancestralProbs = NULL) 
  {
    if (is.data.frame(X)) {
      X <- as.matrix(X)
    }
    if (is.data.frame(Y)) {
      Y <- as.matrix(Y)
    }
    if (nrow(X) == 1 || nrow(X) < minPointsForSplit || depth > 
        maxDepthSplit) {
      tree <- setupLeaf(Y)     ###### CHANGED
      class(tree) <- "canonical_correlation_tree"     ###### CHANGED
      return(tree)     ###### CHANGED
      ######return(setupLeaf(Y))
    }
    else if (is.matrix(Y) && ncol(Y) > 1) {
      if (sum(abs(colSums(Y)) > 1e-12) == 1) {
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
        ######return(setupLeaf(Y))
      }
    }
    else if (all(Y == 0)) {
      tree <- setupLeaf(Y)     ###### CHANGED
      class(tree) <- "canonical_correlation_tree"     ###### CHANGED
      return(tree)     ###### CHANGED
      ######return(setupLeaf(Y))
    }
    if (nrow(X) == 2) {
      if (all(Y[1, ] == Y[2, ])) {
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
        ######return(setupLeaf(Y))
      }
      projection_matrix = t(X[2, , drop = F] - X[1, , drop = F])
      partitionPoint = 0.5 * (X[2, ] %*% projection_matrix + 
                                X[1, ] %*% projection_matrix)
      lessThanPartPoint = (X %*% projection_matrix) <= partitionPoint[1]
      if (lessThanPartPoint[1]==lessThanPartPoint[2]) {     ###### CHANGED
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
      }     ###### CHANGED
      best_split = list(partitionPoint = partitionPoint[1], 
                        splitDir = 1, lessThanPartPoint = lessThanPartPoint)
    }
    else {
      if (projectionBootstrap) {
        sampleIndices = sample(nrow(X), size = nrow(X), replace = T)
        XSampled = X[sampleIndices, , drop = F]
        YSampled = Y[sampleIndices, , drop = F]
        if (sum(abs(colSums(YSampled)) > 1e-12) == 1) {
          tree <- setupLeaf(Y)     ###### CHANGED
          class(tree) <- "canonical_correlation_tree"     ###### CHANGED
          return(tree)     ###### CHANGED
          ######return(setupLeaf(Y))
        }
        cca = canonical_correlation_analysis(XSampled, YSampled)
      }
      else {
        cca = canonical_correlation_analysis(X, Y)
      }
      projection_matrix = cca$xcoef
      U = X %*% projection_matrix
      bUVaries = apply(U, 2, stats::var) > xVariationTolerance
      if (!any(bUVaries)) {
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
        ######return(setupLeaf(Y))
      }
      best_split = find_best_split(U[, bUVaries, drop = F], 
                                   Y, xVariationTolerance)
      if (best_split$gain < 0) {
        tree = setupLeaf(Y)
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)
      }
    }
    countsNode = colSums(Y)
    nonZeroCounts = sum(countsNode > 0)
    uniqueNonZeroCounts = length(unique(countsNode)[unique(countsNode) != 
                                                      0])
    if (uniqueNonZeroCounts == nonZeroCounts || is.null(ancestralProbs)) {
      ancestralProbs = countsNode/sum(countsNode)
    }
    else {
      ancestralProbs = rbind(ancestralProbs, countsNode/sum(countsNode))
    }
    treeLeft = canonical_correlation_tree(X[best_split$lessThanPartPoint, 
                                            , drop = F], Y[best_split$lessThanPartPoint, , drop = F], 
                                          depth = depth + 1, minPointsForSplit = minPointsForSplit, 
                                          maxDepthSplit = maxDepthSplit, ancestralProbs = ancestralProbs)
    treeRight = canonical_correlation_tree(X[!best_split$lessThanPartPoint, 
                                             , drop = F], Y[!best_split$lessThanPartPoint, , drop = F], 
                                           depth = depth + 1, minPointsForSplit = minPointsForSplit, 
                                           maxDepthSplit = maxDepthSplit, ancestralProbs = ancestralProbs)
    model = structure(list(isLeaf = F, trainingCounts = countsNode, 
                           decisionProjection = projection_matrix[, best_split$splitDir, 
                                                                  drop = FALSE], partitionPoint = best_split$partitionPoint, 
                           depth = depth, refLeftChild = treeLeft, refRightChild = treeRight), 
                      class = "canonical_correlation_tree")
    return(model)
  }
  
  reassignInPackage(name="canonical_correlation_tree", pkgName="ccf", value=mycanonical_correlation_tree)
  
  
  
  # Dummy code categorical variables:  
  require("fastDummies")

  factorvars <- setdiff(names(data)[sapply(data, is.factor)], dependent.variable.name)
  if (length(factorvars) > 0) {
    datatrans <- fastDummies::dummy_cols(data, select_columns = factorvars, remove_selected_columns=TRUE, remove_first_dummy=TRUE)
  } else {
    datatrans <- data 
  }
  
  # Construct canonical correlation forest:
  model <- ccf::canonical_correlation_forest(x=as.matrix(datatrans[,names(datatrans)!=dependent.variable.name]), y=datatrans[,dependent.variable.name], ...)
  
  return(model)
  
}





# Prediction function for Canonical Correlation Forest:

predictCanForest <- function(model, dependent.variable.name, newdata) {
  
  
  # This is an altered version of the function setupLeaf from
  # the ccf package.
  # The original version of the package was not able to deal
  # with situations, in which both classes in leafs were
  # represented by the same numbers of observations.
  # In the altered version, one of the two classes is
  # assigned randomly in these situations.
  
  mysetupLeaf <- function (y) 
  {
    if (is.matrix(y)) {
      countsNode <- colSums(y)
      classIndex <- nnet::which.is.max(countsNode)
      #### maxCounts <- max(countsNode)
      #### equalMaxCounts <- maxCounts == countsNode
      #### if (sum(equalMaxCounts) == 1) {
      ####     classIndex <- which(equalMaxCounts)
      #### }
      #### else {
      ####     TODO("TODO multiple max tie breaking")
      #### }
    }
    else {
      countsNode <- sum(y)
      classIndex <- 1
    }
    return(list(isLeaf = TRUE, classIndex = classIndex, trainingCounts = countsNode))
  }
  
  library("ccf")
  library("R.utils")
  reassignInPackage(name="setupLeaf", pkgName="ccf", value=mysetupLeaf)
  
  
  
  
  # This is an altered version of the function canonical_correlation_tree from
  # the ccf package.
  
  # The following issues were fixed here:
  # 1. In situations in which a tree did not involve at least one split, the class
  #    "canonical_correlation_tree" was not assigned to the tree, which lead to an error
  #    when applying the prediction function, because the latter needs the
  #    trees to have class "canonical_correlation_tree".
  # 2. For nodes, which featured only two observations ("if (nrow(X) == 2) {"),
  #    which did not feature the same class, it was possible that the
  #    training of the tree ran into an infinite loop.
  
  
  mycanonical_correlation_tree <- function (X, Y, depth = 0, minPointsForSplit = 2, maxDepthSplit = Inf, 
                                            xVariationTolerance = 1e-10, projectionBootstrap = FALSE, 
                                            ancestralProbs = NULL) 
  {
    if (is.data.frame(X)) {
      X <- as.matrix(X)
    }
    if (is.data.frame(Y)) {
      Y <- as.matrix(Y)
    }
    if (nrow(X) == 1 || nrow(X) < minPointsForSplit || depth > 
        maxDepthSplit) {
      tree <- setupLeaf(Y)     ###### CHANGED
      class(tree) <- "canonical_correlation_tree"     ###### CHANGED
      return(tree)     ###### CHANGED
      ######return(setupLeaf(Y))
    }
    else if (is.matrix(Y) && ncol(Y) > 1) {
      if (sum(abs(colSums(Y)) > 1e-12) == 1) {
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
        ######return(setupLeaf(Y))
      }
    }
    else if (all(Y == 0)) {
      tree <- setupLeaf(Y)     ###### CHANGED
      class(tree) <- "canonical_correlation_tree"     ###### CHANGED
      return(tree)     ###### CHANGED
      ######return(setupLeaf(Y))
    }
    if (nrow(X) == 2) {
      if (all(Y[1, ] == Y[2, ])) {
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
        ######return(setupLeaf(Y))
      }
      projection_matrix = t(X[2, , drop = F] - X[1, , drop = F])
      partitionPoint = 0.5 * (X[2, ] %*% projection_matrix + 
                                X[1, ] %*% projection_matrix)
      lessThanPartPoint = (X %*% projection_matrix) <= partitionPoint[1]
      if (lessThanPartPoint[1]==lessThanPartPoint[2]) {     ###### CHANGED
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
      }     ###### CHANGED
      best_split = list(partitionPoint = partitionPoint[1], 
                        splitDir = 1, lessThanPartPoint = lessThanPartPoint)
    }
    else {
      if (projectionBootstrap) {
        sampleIndices = sample(nrow(X), size = nrow(X), replace = T)
        XSampled = X[sampleIndices, , drop = F]
        YSampled = Y[sampleIndices, , drop = F]
        if (sum(abs(colSums(YSampled)) > 1e-12) == 1) {
          tree <- setupLeaf(Y)     ###### CHANGED
          class(tree) <- "canonical_correlation_tree"     ###### CHANGED
          return(tree)     ###### CHANGED
          ######return(setupLeaf(Y))
        }
        cca = canonical_correlation_analysis(XSampled, YSampled)
      }
      else {
        cca = canonical_correlation_analysis(X, Y)
      }
      projection_matrix = cca$xcoef
      U = X %*% projection_matrix
      bUVaries = apply(U, 2, stats::var) > xVariationTolerance
      if (!any(bUVaries)) {
        tree <- setupLeaf(Y)     ###### CHANGED
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)     ###### CHANGED
        ######return(setupLeaf(Y))
      }
      best_split = find_best_split(U[, bUVaries, drop = F], 
                                   Y, xVariationTolerance)
      if (best_split$gain < 0) {
        tree = setupLeaf(Y)
        class(tree) <- "canonical_correlation_tree"     ###### CHANGED
        return(tree)
      }
    }
    countsNode = colSums(Y)
    nonZeroCounts = sum(countsNode > 0)
    uniqueNonZeroCounts = length(unique(countsNode)[unique(countsNode) != 
                                                      0])
    if (uniqueNonZeroCounts == nonZeroCounts || is.null(ancestralProbs)) {
      ancestralProbs = countsNode/sum(countsNode)
    }
    else {
      ancestralProbs = rbind(ancestralProbs, countsNode/sum(countsNode))
    }
    treeLeft = canonical_correlation_tree(X[best_split$lessThanPartPoint, 
                                            , drop = F], Y[best_split$lessThanPartPoint, , drop = F], 
                                          depth = depth + 1, minPointsForSplit = minPointsForSplit, 
                                          maxDepthSplit = maxDepthSplit, ancestralProbs = ancestralProbs)
    treeRight = canonical_correlation_tree(X[!best_split$lessThanPartPoint, 
                                             , drop = F], Y[!best_split$lessThanPartPoint, , drop = F], 
                                           depth = depth + 1, minPointsForSplit = minPointsForSplit, 
                                           maxDepthSplit = maxDepthSplit, ancestralProbs = ancestralProbs)
    model = structure(list(isLeaf = F, trainingCounts = countsNode, 
                           decisionProjection = projection_matrix[, best_split$splitDir, 
                                                                  drop = FALSE], partitionPoint = best_split$partitionPoint, 
                           depth = depth, refLeftChild = treeLeft, refRightChild = treeRight), 
                      class = "canonical_correlation_tree")
    return(model)
  }
  
  reassignInPackage(name="canonical_correlation_tree", pkgName="ccf", value=mycanonical_correlation_tree)
  
  
  
  # This is a modified version of the the original predict.canonical_correlation_forest function.
  # The latter function did only deliver point predictions, but no predictions of the
  # class probabilities.
  # Therefore, we altered the original function to feature this functionality:
  # Here, we simply use the proportions of the trees that predict the classes
  # as predicted class probabilities.
  
  mycanpredict <- function (object, newdata, verbose = FALSE, ...) 
  {
    
    if (missing(newdata)) {
      stop("Argument 'newdata' is missing.")
    }
    ntree <- length(object$forest)
    treePredictions <- matrix(NA, nrow = nrow(newdata), ncol = ntree)
    if (verbose) {
      cat("calculating predictions\n")
    }
    treePredictions = lapply(object$forest, predict, newdata)
    treePredictions = do.call(cbind, treePredictions)
    
    # Proportion of trees predicting class 2:
    probpreds <- apply(treePredictions, 1, function(x) mean(x==2))
    
    if (verbose) {
      cat("Majority vote\n")
    }
    
    # Majority voting (not altered):
    classpreds <- apply(treePredictions, 1, function(row) {
      names(which.max(table(row)))
    })
    return(list(probpreds=probpreds, classpreds=classpreds))
  }
  
  environment(mycanpredict) <- environment(canonical_correlation_forest)
  
  
  
  
  require("fastDummies")

  # Dummy code factor variables:
  factorvars <- setdiff(names(newdata)[sapply(newdata, is.factor)], dependent.variable.name)
  if (length(factorvars) > 0) {
    newdatatrans <- fastDummies::dummy_cols(newdata, select_columns = factorvars, remove_selected_columns=TRUE, remove_first_dummy=TRUE)
  } else {
    newdatatrans <- newdata 
  }
  
  # Obtain predictions:
  canpreds <- mycanpredict(object=model, newdata=as.matrix(newdatatrans[,names(newdatatrans)!=dependent.variable.name]))
  probpreds <- canpreds$probpreds
  classpreds <- factor(levels(newdata[,dependent.variable.name])[as.numeric(canpreds$classpreds)], levels=levels(newdata[,dependent.variable.name]))
  
  return(list(probpreds=probpreds, classpreds=classpreds))
  
}







# Function that constructs a Random Forest:

trainRandomForest <- function(dependent.variable.name, data, ...) {
  
  require("ranger")
  
  # Construct random forest (where the used configuration is adjusted to the default configuration of interaction forest):
  model <- ranger::ranger(dependent.variable.name = dependent.variable.name, data = data, importance="none", replace=FALSE, sample.fraction=0.7, num.tree=2000, probability=TRUE, min.node.size = 5, respect.unordered.factors="order", num.threads=1)
  
  return(model)
  
}


# Prediction function for Random Forest:

predictRandomForest <- function(model, dependent.variable.name, newdata) {
  
  require("ranger")
  
  # Predict conditional class probabilities for the observations in
  # the test data set:
  probpreds <- predict(model, data=newdata)$predictions[,2]
  
  # Obtain class predictions:
  classpreds <- factor(ifelse(probpreds > 0.5, levels(newdata[,dependent.variable.name])[2], levels(newdata[,dependent.variable.name])[1]), levels=levels(newdata[,dependent.variable.name]))
  if(any(probpreds==0.5))
    classpreds[probpreds==0.5] <- sample(levels(newdata[,dependent.variable.name]), size=sum(probpreds==0.5), replace=TRUE)
  
  return(list(probpreds=probpreds, classpreds=classpreds))
  
}







# Function that constructs an Interaction Forest:

trainInteractionForest <- function(dependent.variable.name, data, ...) {
  
  require("diversityForest")
  
  # Construct Interaction Forest:
  model <- diversityForest::interactionfor(dependent.variable.name = dependent.variable.name, data = data, importance="none", probability=TRUE, num.threads=1)
  
  return(model)
  
}


# Prediction function for Interaction Forest:

predictInteractionForest <- function(model, dependent.variable.name, newdata) {
  
  require("diversityForest")
  
  # Predict conditional class probabilities for the observations in
  # the test data set:
  probpreds <- predict(model, data=newdata)$predictions[,2]
  
  # Obtain class predictions:
  classpreds <- factor(ifelse(probpreds > 0.5, levels(newdata[,dependent.variable.name])[2], levels(newdata[,dependent.variable.name])[1]), levels=levels(newdata[,dependent.variable.name]))
  if(any(probpreds==0.5))
    classpreds[probpreds==0.5] <- sample(levels(newdata[,dependent.variable.name]), size=sum(probpreds==0.5), replace=TRUE)
  
  return(list(probpreds=probpreds, classpreds=classpreds))
  
}
















# Function calculating AUC values (taken from "mlr" R package):

measureAUC <- function (probabilities, truth, negative, positive) 
{
  if (is.factor(truth)) {
    i = as.integer(truth) == which(levels(truth) == positive)
  }
  else {
    i = truth == positive
  }
  if (length(unique(i)) < 2L) {
    stop("truth vector must have at least two classes")
  }
  if (length(i) > 5000L) {
    r = frankv(probabilities)
  }
  else {
    r = rank(probabilities)
  }
  n.pos = as.numeric(sum(i))
  n.neg = length(i) - n.pos
  (sum(r[i]) - n.pos * (n.pos + 1)/2)/(n.pos * n.neg)
}




# Function calculating ACC values (taken from "mlr" R package):

measureACC <- function (truth, response) 
{
  mean(response == truth)
}



# Function calculating Brier values (taken from "mlr" R package):

measureBrier <- function (probabilities, truth, negative, positive) 
{
  y = as.numeric(truth == positive)
  mean((y - probabilities)^2)
}







# Function that generates the splittings for stratified cross-validation:

# Input parameters:

# data   - data set
# yname  - label of the target variable
# ncv    - number of folds to use

makeCVdiv <- function(data, yname="y", ncv=5) {
  
  cvid <- rep(NA, length = nrow(data))
  for(i in levels(data[, yname])) cvid[data[, yname] == i] <- sample(rep(1:ncv, length = sum(data[, yname] == i)))
  
  cvid
  
}
