# This function generates a simulated data set and calculates the importance values for
# all methods except for the IMDMS.
#
# It takes the whole number 'iter', which corresponds to the iter-th line 
# of 'scenariogrid', which contains the necessary information
# on the iter-th setting.

evaluatesetting <- function(iter) {
  
  # Obtain information for the iter-th setting:
  
  n <- scenariogrid$n[iter]
  itind <- scenariogrid$itind[iter]
  seed <- scenariogrid$seed[iter]
  settingid <- scenariogrid$settingid[iter]
  
  
  # Set seed:
  
  set.seed(seed)
  
  
  # Simulate data set:
  
  dataset <- simDataset(N=n, mainstrength=c(3, 2, 1), quantstrength=c(3, 2, 1), qualstrength=c(3, 2, 1))
  
  
  # EIM:
  
  require("diversityForest")
  model <- interactionfor(dependent.variable.name = "y", data = dataset, num.trees = 20000, importance="both", probability=TRUE, num.threads=1)
  eim.univ <- model$eim.univ
  eim.quant <- model$eim.quant
  eim.qual <- model$eim.qual
  
  # RF-V-pairs:
  
  require("ranger")
  model <- ranger(dependent.variable.name = "y", data = dataset, replace=FALSE, sample.fraction=0.7, num.trees = 20000, importance="permutation", probability=TRUE, min.node.size = 5, num.threads=1)
  variable.importance <- model$variable.importance
  combs <- combn(1:(ncol(dataset)-1), 2)
  variable.importance.biv <- apply(combs, 2, function(x) (variable.importance[x[1]] + variable.importance[x[2]])/2)
  names(variable.importance.biv) <- apply(combs, 2, function(x) paste(names(variable.importance)[x[1]], names(variable.importance)[x[2]], sep=" AND "))
  
  # iRF:
  
  require("iRF")
  iRF.biv <- iRF(x=as.matrix(dataset[,names(dataset)!="y"]), y=dataset$y, interactions.return=5, ntree = 20000)$interaction[[5]]
  
  
  # PA:
  
  require("randomForestSRC")
  SRC_obj <- rfsrc(y ~., data=dataset, importance=TRUE, ntree = 20000)
  SRC.biv <- find.interaction(SRC_obj, method="vimp", verbose=FALSE)[,"Difference"]
  newnames <- sapply(names(SRC.biv), function(x) paste(sort(strsplit(x, split=":")[[1]]), collapse=" AND "))
  names(SRC.biv) <- newnames
  
  
  # Combine results:
  
  res <- list(eim.univ=eim.univ, eim.quant=eim.quant, eim.qual=eim.qual,
              variable.importance=variable.importance, variable.importance.biv=variable.importance.biv,
              iRF.biv=iRF.biv, SRC.biv=SRC.biv)
  
  # Return results:
  
  return(res)
  
}







# This function generates a simulated data set and calculates the importance values for the IMDMS.
#
# It takes the whole number 'iter', which corresponds to the iter-th line 
# of 'scenariogrid', which contains the necessary information
# on the iter-th setting.

evaluatesetting2 <- function(iter) {
  
  # Obtain information for the iter-th setting:
  
  n <- scenariogrid$n[iter]
  itind <- scenariogrid$itind[iter]
  seed <- scenariogrid$seed[iter]
  settingid <- scenariogrid$settingid[iter]
  
  
  # Set seed:
  
  set.seed(seed)
  
  
  # Simulate data set:
  
  dataset <- simDataset(N=n, mainstrength=c(3, 2, 1), quantstrength=c(3, 2, 1), qualstrength=c(3, 2, 1))
  
  
  # IMDMS:
  
  require("randomForestSRC")
  SRC_obj <- rfsrc(y ~., data=dataset, importance=TRUE, ntree=20000)
  intmat <- find.interaction(SRC_obj, method="maxsubtree", verbose=FALSE)
  combs <- combn(1:ncol(intmat), 2)
  SRCsubtree.biv <- apply(combs, 2, function(x) min(c(intmat[x[1], x[2]], intmat[x[2], x[1]])))
  names(SRCsubtree.biv) <- apply(combs, 2, function(x) paste(rownames(intmat)[x], collapse=":"))
  newnames <- sapply(names(SRCsubtree.biv), function(x) paste(sort(strsplit(x, split=":")[[1]]), collapse=" AND "))
  names(SRCsubtree.biv) <- newnames
  
  res <- list(SRCsubtree.biv=SRCsubtree.biv)
  
  # Return results:
  
  return(res)
  
}







# This function generates a simulated data set and calculates the importance values for all methods.
# It was used for all settings with n = 1000.
#
# It takes the whole number 'iter', which corresponds to the iter-th line 
# of 'scenariogrid', which contains the necessary information
# on the iter-th setting.

evaluatesettingn1000 <- function(iter) {
  
  # Obtain information for the iter-th setting:
  
  n <- scenariogrid$n[iter]
  itind <- scenariogrid$itind[iter]
  seed <- scenariogrid$seed[iter]
  settingid <- scenariogrid$settingid[iter]
  
  
  # Set seed:
  
  set.seed(seed)
  
  
  # Simulate data set:
  
  dataset <- simDataset(N=n, mainstrength=c(3, 2, 1), quantstrength=c(3, 2, 1), qualstrength=c(3, 2, 1))
  
  
  # EIM:
  
  require("diversityForest")
  model <- interactionfor(dependent.variable.name = "y", data = dataset, num.trees = 20000, importance="both", probability=TRUE, num.threads=1)
  eim.univ <- model$eim.univ
  eim.quant <- model$eim.quant
  eim.qual <- model$eim.qual
  
  
  # RF-V-pairs:
  
  require("ranger")
  model <- ranger(dependent.variable.name = "y", data = dataset, replace=FALSE, sample.fraction=0.7, num.trees = 20000, importance="permutation", probability=TRUE, min.node.size = 5, num.threads=1)
  variable.importance <- model$variable.importance
  combs <- combn(1:(ncol(dataset)-1), 2)
  variable.importance.biv <- apply(combs, 2, function(x) (variable.importance[x[1]] + variable.importance[x[2]])/2)
  names(variable.importance.biv) <- apply(combs, 2, function(x) paste(names(variable.importance)[x[1]], names(variable.importance)[x[2]], sep=" AND "))
  
  
  # iRF:
  
  require("iRF")
  iRF.biv <- iRF(x=as.matrix(dataset[,names(dataset)!="y"]), y=dataset$y, interactions.return=5, ntree = 500)$interaction[[5]]
  
  
  # PA:
  
  require("randomForestSRC")
  SRC_obj <- rfsrc(y ~., data=dataset, importance=TRUE, ntree = 1000)
  SRC.biv <- find.interaction(SRC_obj, method="vimp", verbose=FALSE)[,"Difference"]
  newnames <- sapply(names(SRC.biv), function(x) paste(sort(strsplit(x, split=":")[[1]]), collapse=" AND "))
  names(SRC.biv) <- newnames
  
  
  # IMDMS:
  
  SRC_obj <- rfsrc(y ~., data=dataset, importance=TRUE, ntree=20000)
  intmat <- find.interaction(SRC_obj, method="maxsubtree", verbose=FALSE)
  combs <- combn(1:ncol(intmat), 2)
  SRCsubtree.biv <- apply(combs, 2, function(x) min(c(intmat[x[1], x[2]], intmat[x[2], x[1]])))
  names(SRCsubtree.biv) <- apply(combs, 2, function(x) paste(rownames(intmat)[x], collapse=":"))
  newnames <- sapply(names(SRCsubtree.biv), function(x) paste(sort(strsplit(x, split=":")[[1]]), collapse=" AND "))
  names(SRCsubtree.biv) <- newnames
  
  
  # Combine results:
  
  res <- list(eim.univ=eim.univ, eim.quant=eim.quant, eim.qual=eim.qual,
              variable.importance=variable.importance, variable.importance.biv=variable.importance.biv,
              iRF.biv=iRF.biv, SRC.biv=SRC.biv, SRCsubtree.biv=SRCsubtree.biv)
  
  # Return results:
  
  return(res)
  
}











# Functions used for simulating the data sets:





# Function for simulating the variables with main effects only:

simMainEff <- function(N, strength) {
  if(strength==1)
    d <- 0.5
  if(strength==2)
    d <- 0.75
  if(strength==3)
    d <- 1
  x <- c(rnorm(N/2, mean=0, sd=1), rnorm(N/2, mean=d, sd=1))
  return(x)
}





# Functions for simulating variable pairs with quantitative interactions:



# Function for drawing a point from class 1, if class 2 should concentrate in lower left corner:

drawQuantInt1cl1 <- function(dqual, propcorner) {
  dind <- sample(1:3, size=1)
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}

# Function for drawing a point from class 2, if class 2 should concentrate in lower left corner:

drawQuantInt1cl2 <- function(dqual, propcorner) {
  dind <- sample(1:4, size=1, prob=c((1-propcorner)/3, (1-propcorner)/3, propcorner, (1-propcorner)/3))
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
  if(dind==4) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}




# Function for drawing a point from class 1, if class 2 should concentrate in lower right corner:

drawQuantInt2cl1 <- function(dqual, propcorner) {
  dind <- sample(1:3, size=1)
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
}

# Function for drawing a point from class 2, if class 2 should concentrate in lower right corner:

drawQuantInt2cl2 <- function(dqual, propcorner) {
  dind <- sample(1:4, size=1, prob=c((1-propcorner)/3, (1-propcorner)/3, (1-propcorner)/3, propcorner))
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
  if(dind==4) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}





# Function for drawing a point from class 1, if class 2 should concentrate in upper right corner:

drawQuantInt3cl1 <- function(dqual, propcorner) {
  dind <- sample(1:3, size=1)
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}

# Function for drawing a point from class 2, if class 2 should concentrate in upper right corner:

drawQuantInt3cl2 <- function(dqual, propcorner) {
  dind <- sample(1:4, size=1, prob=c((1-propcorner)/3, propcorner, (1-propcorner)/3, (1-propcorner)/3))
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
  if(dind==4) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}





# Function for drawing a point from class 1, if class 2 should concentrate in upper left corner:

drawQuantInt4cl1 <- function(dqual, propcorner) {
  dind <- sample(1:3, size=1)
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}

# Function for drawing a point from class 2, if class 2 should concentrate in upper left corner:

drawQuantInt4cl2 <- function(dqual, propcorner) {
  dind <- sample(1:4, size=1, prob=c(propcorner, (1-propcorner)/3, (1-propcorner)/3, (1-propcorner)/3))
  if(dind==1) {
    return(rmvnorm(n=1, mean=c(0,dqual))[1,])
  }
  if(dind==2) {
    return(rmvnorm(n=1, mean=c(dqual,dqual))[1,])
  }
  if(dind==3) {
    return(rmvnorm(n=1, mean=c(0,0))[1,])
  }
  if(dind==4) {
    return(rmvnorm(n=1, mean=c(dqual,0))[1,])
  }
}




# Function for simulating a pair of variables with quantitative interaction:

simQuantInt <- function(N, corner, strength) {
  
  # Set parameters for specified effect strength level:  
  if(strength==1) {
    dqual <- 1.225
    propcorner <- 0.485
  }
  if(strength==2) {
    dqual <- 1.51
    propcorner <- 0.574
  }
  if(strength==3) {
    dqual <- 1.772
    propcorner <- 0.649
  }
  
  x1 <- x2 <- rep(0, N)
  
  # If class 2 in lower left corner:
  if(corner == 1) {
    draws1 <- replicate(N/2, drawQuantInt1cl1(dqual=dqual, propcorner=propcorner))
    draws2 <- replicate(N/2, drawQuantInt1cl2(dqual=dqual, propcorner=propcorner))
  }
  
  # If class 2 in lower right corner:
  if(corner == 2) {
    draws1 <- replicate(N/2, drawQuantInt2cl1(dqual=dqual, propcorner=propcorner))
    draws2 <- replicate(N/2, drawQuantInt2cl2(dqual=dqual, propcorner=propcorner))
  }
  
  # If class 2 in upper right corner:
  if(corner == 3) {
    draws1 <- replicate(N/2, drawQuantInt3cl1(dqual=dqual, propcorner=propcorner))
    draws2 <- replicate(N/2, drawQuantInt3cl2(dqual=dqual, propcorner=propcorner))
  }
  
  # If class 2 in upper left corner:
  if(corner == 4) {
    draws1 <- replicate(N/2, drawQuantInt4cl1(dqual=dqual, propcorner=propcorner))
    draws2 <- replicate(N/2, drawQuantInt4cl2(dqual=dqual, propcorner=propcorner))
  }
  
  # Assign drawn values to the variables:
  x1[1:(N/2)] <- draws1[1,]
  x2[1:(N/2)] <- draws1[2,]
  x1[((N/2)+1):N] <- draws2[1,]
  x2[((N/2)+1):N] <- draws2[2,]
  
  # Return variables:
  return(cbind(x1, x2))
}









# Functions for simulating variable pairs with qualitative interactions:



# Function for drawing a value in upper right or lower left corner:

drawQualInt1 <- function(dqual) {
  dind <- sample(1:2, size=1)
  if(dind==1) {
    x2_1 <- rnorm(1, mean=dqual, sd=1)
    x2_2 <- rnorm(1, mean=dqual, sd=1)
  }
  else {
    x2_1 <- rnorm(1, mean=0, sd=1)
    x2_2 <- rnorm(1, mean=0, sd=1)
  }
  return(c(x2_1, x2_2))
}


# Function for drawing a value in upper left or lower right corner:

drawQualInt2 <- function(dqual) {
  dind <- sample(1:2, size=1)
  if(dind==1) {
    x2_1 <- rnorm(1, mean=0, sd=1)
    x2_2 <- rnorm(1, mean=dqual, sd=1)
  }
  else {
    x2_1 <- rnorm(1, mean=dqual, sd=1)
    x2_2 <- rnorm(1, mean=0, sd=1)
  }
  return(c(x2_1, x2_2))
}




# Function for simulating a pair of variables with qualitative interaction:

simQualInt <- function(N, corner, strength) {
  
  # Set parameters for specified effect strength level:  
  if(strength==1)
    dqual <- 1.225
  if(strength==2)
    dqual <- 1.51
  if(strength==3)
    dqual <- 1.772
  
  x1 <- x2 <- rep(0, N)
  
  # If class 2 in upper left and lower right corner:
  if(corner == 1) {
    
    draws <- replicate(N/2, drawQualInt1(dqual=dqual))
    x1[1:(N/2)] <- draws[1,]
    x2[1:(N/2)] <- draws[2,]
    
    draws <- replicate(N/2, drawQualInt2(dqual=dqual))
    x1[((N/2)+1):N] <- draws[1,]
    x2[((N/2)+1):N] <- draws[2,]
  }
  
  # If class 2 in upper right and lower left corner:
  if(corner == 2) {
    
    draws <- replicate(N/2, drawQualInt2(dqual=dqual))
    x1[1:(N/2)] <- draws[1,]
    x2[1:(N/2)] <- draws[2,]
    
    draws <- replicate(N/2, drawQualInt1(dqual=dqual))
    x1[((N/2)+1):N] <- draws[1,]
    x2[((N/2)+1):N] <- draws[2,]
  }
  
  # Return variables:
  return(cbind(x1, x2))
}






# Function for simulating a data set:

simDataset <- function(N, mainstrength, quantstrength, qualstrength) {
  
  require("mvtnorm")
  
  # Empty data frame:
  simdata <- data.frame(matrix(nrow=N, ncol=68))
  
  # Add response variable:
  simdata$y <- factor(rep(1:2, each=N/2), levels=1:2)
  
  # Add variables without effect:
  count <- 1
  for(i in 1:50) {
    simdata[,count] <- rnorm(N, mean=0, sd=1)
    count <- count + 1
  }
  
  # Add variables with main effects only:
  simdata[,count] <- simMainEff(N, strength=mainstrength[1])
  count <- count + 1
  simdata[,count] <- simMainEff(N, strength=mainstrength[1])
  count <- count + 1
  
  simdata[,count] <- simMainEff(N, strength=mainstrength[2])
  count <- count + 1
  simdata[,count] <- simMainEff(N, strength=mainstrength[2])
  count <- count + 1
  
  simdata[,count] <- simMainEff(N, strength=mainstrength[3])
  count <- count + 1
  simdata[,count] <- simMainEff(N, strength=mainstrength[3])
  count <- count + 1
  
  
  # Add variables with quantitative interactions:
  simdata[,count + c(0,1)] <- simQuantInt(N, corner=2, strength=quantstrength[1])
  count <- count + 2
  
  simdata[,count + c(0,1)] <- simQuantInt(N, corner=3, strength=quantstrength[2])
  count <- count + 2
  
  simdata[,count + c(0,1)] <- simQuantInt(N, corner=4, strength=quantstrength[3])
  count <- count + 2
  
  
  # Add variables with qualitative interactions:  
  simdata[,count + c(0,1)] <- simQualInt(N, corner=2, strength=qualstrength[1])
  count <- count + 2
  
  simdata[,count + c(0,1)] <- simQualInt(N, corner=1, strength=qualstrength[2])
  count <- count + 2
  
  simdata[,count + c(0,1)] <- simQualInt(N, corner=2, strength=qualstrength[3])
  count <- count + 2
  
  
  # Return data set:
  return(simdata)
  
}
